"""Identitying catalytic residues in protein structures by matching catalytic templates
    derived from the Mechanism and Catalytic Site Atlas

See Also:
    *PAPER TITLE*
    Authors
    bioRxiv TAG
    `doi:goes_here <link_goes_here>`_

"""

__author__ = "Raymund Hackett"
__license__ = "MIT"
__version__ = "0.1.2"
